package jp.co.sds_corp.test;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import junit.framework.AssertionFailedError;

import org.apache.log4j.Logger;
import org.seasar.extension.dataset.ColumnType;
import org.seasar.extension.dataset.DataRow;
import org.seasar.extension.dataset.DataSet;
import org.seasar.extension.dataset.DataTable;
import org.seasar.extension.dataset.types.ColumnTypes;
import org.seasar.extension.unit.S2TestCase;
import org.seasar.framework.container.ComponentDef;
import org.seasar.framework.container.S2Container;
import org.seasar.framework.container.factory.S2ContainerFactory;
import org.seasar.framework.util.StringUtil;

public abstract class BasicTestCase extends S2TestCase
{
	protected Logger logger = Logger.getLogger(getClass());
	Date start;

	/**
	 * j2eeに関する設定を単体テスト実行プロセス単位で1つ用意しておくための
	 * 設定ファイル名.<br/>
	 */
	public static final String defaultJ2eeDicon = "default_j2ee_ut.dicon";
	/**
	 * j2eeに関する設定をテストメソッド単位で用意するための設定ファイル名.<br/>
	 * 通常は、defaultJ2eeDiconで保持しているコンポーネントがコピーされる。
	 */
	public static final String j2eeDicon = "j2ee_ut.dicon";

	static S2Container defaultS2Container;
	static List<ComponentDef> defaultComponentDefList;
	
	static {
		defaultS2Container = S2ContainerFactory.create(defaultJ2eeDicon);
		defaultComponentDefList = createComponentDefList(defaultS2Container);
	}

	 /**
	 * コンポーネント定義リスト生成
	 * <br/>
	 * <br/>
	 * {@link ComponentDef}の{@link List}を生成する。
	 * <br/>
	 * @param container　{@link S2Container}
	 * @return containerに含まれている取得可能なコンポーネント定義のリスト。
	 */
	static List<ComponentDef> createComponentDefList(final S2Container container) {
		Logger logger = Logger.getLogger(BasicTestCase.class);
		assert container != null;
		final List<ComponentDef> list =  new ArrayList<ComponentDef>();
		final int size = container.getComponentDefSize();
		for (int i = 0; i < size; ++i) {
			final ComponentDef componentDef = container.getComponentDef(i);
			final String name = componentDef.getComponentName();
			final Object component;
			try {
				component = componentDef.getComponent();
			} catch (final RuntimeException ex) {
				logger.info("    ignore component, index: "
						+ i + ", name: " + name, ex);
				continue;
			}
			
			logger.trace("    resolved component,  name: "
					+ name + " as " + component.getClass().getSimpleName());
			list.add(componentDef);
		}
		
		return list;
	}

	@Override
	protected void setUpAfterContainerInit() {
		logger.trace("BEGIN setUpAfterContainerInit()");
		for (final ComponentDef def: defaultComponentDefList) {
			final String name = def.getComponentName();
			final Object component = def.getComponent();
			logger.trace(
				"    register component  name: "
				+ name
				+ " as " + component.getClass().getSimpleName());
			if (StringUtil.isEmpty(name)) {
				register(component);
			} else {
				register(component, name);
			}
		}
		logger.trace("END setUpAfterContainerInit()");
	}

    @Override
	protected void setUp() throws Exception {
		super.setUp();
        include(j2eeDicon);

		logger.debug("=== START " + getClass().getName() + " " + getName() + "()");
		start = new Date();
	}

	@Override
	protected void tearDown() throws Exception {
		if (start == null) {
			logger.warn("########### call super.setUp().");
		}
		performLog(start, "=== END " + getClass().getName() + " " + getName() + "()");
		super.tearDown();
	}

	protected Date performLog(Date start, String msg) {
		Date now = new Date();
		if (start == null) {
			logger.debug(msg);
		} else {
			logger.debug(msg + " " + (now.getTime() - start.getTime()) + "ms.");
		}
		return now;
	}

    /**
     * データベースを初期化する。<br/>
     * この際に単体テストクラスの完全修飾名の"."(ドット)を"/"に変換した
     * ディレクトリ配下の&lt;テストメソッド名&gt;+"_init.xls"ファイルを
     * データベースに書き込む。<br/>
     * 更に"default.xls"ファイルがトップディレクトリにある場合は
     * そのファイルを用いて初期化を行う。<br/>
     * <ul>初期化の順序
     * <li>default.xlsによる完全な置き換え
     * (xlsにないデータはテーブルから削除される)</li>
     * <li>&lt;テストメソッド名&gt;+"_init.xls"ファイルによる置き換え
     * (xlsにファイルがなくても削除されない)</li>
     * </ul>
     */
    protected void loadDataBaseWithDefault() {
        loadDataBaseWithDefault("default.xls");
    }

    /**
     * データベースを初期化する。<br/>
     * この際に単体テストクラスの完全修飾名の"."(ドット)を"/"に変換した
     * ディレクトリ配下の&lt;テストメソッド名&gt;+"_init.xls"ファイルを
     * データベースに書き込む。<br/>
     * 更にdefaultNameファイルがトップディレクトリにある場合は
     * そのファイルを用いて初期化を行う。<br/>
     * <ul>初期化の順序
     * <li>defaultNameによる完全な置き換え
     * (xlsにないデータはテーブルから削除される)</li>
     * <li>&lt;テストメソッド名&gt;+"_init.xls"ファイルによる置き換え
     * (xlsにファイルがなくても削除されない)</li>
     * </ul>
     * @param defaultName デフォルトの初期化用のxlsファイル名
     */
    void loadDataBaseWithDefault(final String defaultName) {
        readXlsAllReplaceDb(defaultName);
        loadDataBase();
    }

    /**
     * データベースを初期化する。<br/>
     * この際に単体テストクラスの完全修飾名の"."(ドット)を"/"に変換した
     * ディレクトリ配下の&lt;テストメソッド名&gt;+"_init.xls"ファイルを
     * データベースに書き込む。<br/>
     * <ul>
     * <li>&lt;テストメソッド名&gt;+"_init.xls"ファイルによる置き換え
     * (xlsにファイルがなくても削除されない)</li>
     * </ul>
     */
    protected void loadDataBase() {
    	loadDataBase(null);
     }

    /**
     * データベースを初期化する。<br/>
     * この際に単体テストクラスの完全修飾名の"."(ドット)を"/"に変換した
     * ディレクトリ配下の&lt;テストメソッド名&gt;+"_init.xls"ファイルを
     * データベースに書き込む。<br/>
     * <ur>
     * <li>&lt;テストメソッド名&gt;+"_init.xls"ファイルによる置き換え
     * (xlsにファイルがなくても削除されない)</li>
     * </ur>
     * @param initName 初期化用のxlsファイル名
     */
    void loadDataBase(final String initName) {
        readXlsReplaceDb(
                getClass().getSimpleName() + "/" + getName() + (initName == null ? "" : "_" + initName) + ".xls");
    }

    /**
     * データベースの中身を検査する。<br/>
     * 単体テストクラスの完全修飾名の"."(ドット)を"/"に変換した
     * ディレクトリ配下の&lt;テストメソッド名&gt;+"_expected.xls"ファイルと
     * マッチしなければ、例外が発生する。
     * @param message 例外メッセージ。
     */
    protected void assertDataBase(final String message) {
        assertDataBase(message, "expected.xls");
    }

    /**
     * データベースの中身を検査する。<br/>
     * 単体テストクラスの完全修飾名の"."(ドット)を"/"に変換した
     * ディレクトリ配下の&lt;テストメソッド名&gt;+"_" + expectedNameファイルと
     * マッチしなければ、例外が発生する。
     * @param message 例外メッセージ。
     */
    void assertDataBase(final String message, final String expectedName) {
        assertDataBase(
                message,
                readXls(
                        getClass().getSimpleName()
                        + "/" +getName() + "_" + expectedName));
    }

    /**
     * {@link DataSet}の中身を検査する。<br/>
     * 単体テストクラスの完全修飾名の"."(ドット)を"/"に変換した
     * ディレクトリ配下の&lt;テストメソッド名&gt;+"_" + expectedNameファイルと
     * マッチしなければ、例外が発生する。
     * @param message 例外メッセージ。
     * @param expected 実行結果。
     */
    protected void assertDataBase(
            final String message, final DataSet expected) {
        final DataSet actual = readDb(expected);
        assertEquals(message, expected, actual);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void assertEquals(
            final String message,
            final DataTable expected,
            final DataTable actual) {
        String outputMessage = message == null ? "" : message;
        outputMessage = outputMessage + ":TableName=" + expected.getTableName();
        assertEquals(
                outputMessage + ":RowSize",
                expected.getRowSize(),
                actual.getRowSize());
        for (int i = 0; i < expected.getRowSize(); ++i) {
            final DataRow expectedRow = expected.getRow(i);
            final DataRow actualRow = actual.getRow(i);
            final List<String> errorMessages = new ArrayList<String>();
            for (int j = 0; j < expected.getColumnSize(); ++j) {
                try {
                    final String columnName = expected.getColumnName(j);
                    final Object expectedValue =
                        expectedRow.getValue(columnName);
                    final ColumnType ct =
                        ColumnTypes.getColumnType(expectedValue);
                    final Object actualValue = actualRow.getValue(columnName);
                    assertColumn(
                            outputMessage + ":Row=" + i
                            + ":columnName=" + columnName,
                            expectedValue,
                            actualValue,
                            ct);
                } catch (final AssertionFailedError e) {
                    errorMessages.add(e.getMessage());
                }
            }
            if (!errorMessages.isEmpty()) {
                fail(message + errorMessages);
            }
        }
    }

    /**
     * カラムごとの条件一致を調べている.<br/>
     * もし、expectedの最初の一文字に"!"が含まれている場合は、
     * それ以後の文字列とactualを比較し、一致すると例外を送出する。<br/>
     * それ以外はS2Unitの仕様に準ずる。<br/>
     * @param message メッセージ.
     * @param expected 予測値.
     * @param actual 実行結果.
     * @param ct {@link ColumnType}
     */
    void assertColumn(
            final String message,
            final Object expected,
            final Object actual,
            final ColumnType ct) {
        if (expected ==null) {
            assertColumnEquals(message, expected, actual, ct);
            return;
        }

        final String expectedString = expected.toString();
        if (StringUtil.isEmpty(expectedString)) {
            assertColumnEquals(message, expected, actual, ct);
            return;
        }

        if (!expectedString.matches("^[!].*$")) {
            assertColumnEquals(message, expected, actual, ct);
            return;
        }
        assertColumnNotEquals(message, expectedString.substring(1), actual, ct);
    }

    /**
     * 一致比較。<br/>
     * @param message メッセージ.
     * @param expected 予測値.
     * @param actual 実行結果.
     * @param ct {@link ColumnType}
     */
    void assertColumnEquals(
            final String message,
            final Object expected,
            final Object actual,
            final ColumnType ct) {
        if (!ct.equals(expected, actual)) {
            assertEquals(message, expected, actual);
        }
    }

    /**
     * 不一致比較。<br/>
     * @param message メッセージ.
     * @param expected 予測値.
     * @param actual 実行結果.
     * @param ct {@link ColumnType}
     */
    void assertColumnNotEquals(
            final String message,
            final Object expected,
            final Object actual,
            final ColumnType ct) {
        assertFalse(message, ct.equals(expected, actual));
        assertEquals(message, expected, actual);
    }

}
